#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>
#include <sys/socket.h>			/* IP headers */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>			/* DNS headers */
#include <net/ethernet.h>		/* Ethernet headers */
#include <netinet/ether.h>
#include <net/if_arp.h>			/* ARP headers */
#include <pcap.h>			/* packet capture header */

#include "yohoo.h"
#include "hash.h"

int verbose = 0;

HASH yohood = NULL;

void yohoo_get_packet(u_char *args, const struct pcap_pkthdr *header,
		const u_char *packet)
{
	char esrc[20], edst[20], asrc[20], adst[20], ipsrc[20], ipdst[20];
	char srchost[80] = "", dsthost[80] ="";
	unsigned long *lptr;
	short op;
	struct in_addr in;
	struct hostent *hostent;
	struct ether_header *eth_head = (struct ether_header *) packet;
	struct arphdr *arp_head =
		(struct arphdr *) (packet + sizeof(struct ether_header));
	struct arp_payload *payload = (struct arp_payload *)
		(packet + sizeof(struct ether_header) +	sizeof(struct arphdr));

	/* the pcap filter should not send us anything that is not ARP,
	   but just in case it fails, we do our own check */
	if (ntohs(eth_head->ether_type) != ETHERTYPE_ARP)
	{
		printf("Ain't an ARP packet, biatch!\n");
		return;
	}

	/* the packet should also be big enough for an ARP message */
	if (header->len < (sizeof(struct ether_header) +
		sizeof(struct arphdr) + sizeof(struct arp_payload)))
	{
		printf("Damn packet is too small, biatch!\n");
		return;
	}

//	printf("\tSource: %s\tDestination: %s\n",
//		ether_ntoa((struct ether_addr *) eth_head->ether_shost),
//		ether_ntoa((struct ether_addr *) eth_head->ether_dhost));

	/* translate Ethernet source and destination into ASCII */
	mac_ntoa(esrc, eth_head->ether_shost);
	mac_ntoa(edst, eth_head->ether_dhost);

	/* translate ARP sender and target into ASCII */
	mac_ntoa(asrc, payload->ar_sha);
	mac_ntoa(adst, payload->ar_tha);

	/* translate IP sender and target into ASCII */
	lptr = (unsigned long *) payload->ar_sip;
	in.s_addr = *lptr;
	strcpy(ipsrc, inet_ntoa(in));
	lptr = (unsigned long *) payload->ar_tip;
	in.s_addr = *lptr;
	strcpy(ipdst, inet_ntoa(in));

	/* finally get the corresponding hostnames */
	hostent = gethostbyaddr(payload->ar_sip, 4, AF_INET);
	if (hostent != NULL)
		strcpy(srchost, hostent->h_name);
	hostent = gethostbyaddr(payload->ar_tip, 4, AF_INET);
	if (hostent != NULL)
		strcpy(dsthost, hostent->h_name);

	/* display a message if Ethernet address differs from ARP */
	if (strcmp(esrc, asrc) != 0)
	{
		printf("WARNING: Ethernet source does not match ARP sender!\n");
		printf("\tSource: %s\tDestination: %s\n", esrc, edst);
		printf("\tType: 0x%04X\n", ntohs(eth_head->ether_type));
	}

	/* other ARP fields that are not too important */
	if (verbose)
	{
		printf("\tHW: 0x%04X (%s) [%d bytes]",
			ntohs(arp_head->ar_hrd), hwtype_ntoa(arp_head->ar_hrd),
			arp_head->ar_hln);
		printf("\tProt: 0x%04X (%s) [%d bytes]\n",
			ntohs(arp_head->ar_pro), prot_ntoa(arp_head->ar_pro),
			arp_head->ar_pln);
	}

	/* print out a message based on ARP operation */
	op = ntohs(arp_head->ar_op);
	switch (op)
	{
		case ARPOP_REQUEST:
			if (verbose)
				printf("%s [%s] looking for %s\n", 
					strlen(srchost) ? srchost : ipsrc, asrc,
					strlen(dsthost) ? dsthost : ipdst);

			add_host(srchost, ipsrc, asrc, 1);
			add_host(dsthost, ipdst, adst, 0);
			break;

		case ARPOP_REPLY:
			if (verbose)
				printf("%s found %s at [%s]\n",
					strlen(dsthost) ? dsthost : ipdst,
					strlen(srchost) ? srchost : ipsrc, asrc);

			add_host(srchost, ipsrc, asrc, 1);
			break;

		default:
			printf("Unknown ARP operation %d\n", op);
			break;
	}
}

void add_host(char *hostname, char *ip, char *mac, int state)
{
	struct host *hostent;
	char *grpname = "Other";
	char *tmpbuf = NULL;

	hostent = hash_hash(yohood, ip);
	assert(hostent);

	/* get the group name */
	if (strlen(hostname) > 0)
	{
		/* strtok() modifies input, so make a copy */
		tmpbuf = malloc(strlen(hostname) + 1);
		assert(tmpbuf);
		strcpy(tmpbuf, hostname);

		strtok(tmpbuf, ".");
		grpname = strtok(NULL, ".");
	}

	if (hostent->tstamp == 0)
	{
		strcpy(hostent->name, hostname);
		strcpy(hostent->group, grpname);
		strcpy(hostent->ip, ip);
		strcpy(hostent->mac, mac);
	}


	if (strcmp(hostent->ip, ip) != 0)
		fprintf(stderr, "WARNING: new \"%s\" same hash as existing \"%s\"\n",
			ip, hostent->ip);

	hostent->tstamp = time(NULL);
	hostent->alive = state;

	if (!tmpbuf) free(tmpbuf);
}

void* yohoo_main(void *arg)
{
	char *dev;				// device
	char errbuf[PCAP_ERRBUF_SIZE];		// error buffer
	pcap_t *handle;				// handle to a PCAP session
	struct bpf_program filter;		// compiled filter code
	char filter_expr[] = "arp";		// ASCII filter expr
	bpf_u_int32 mask;			// netmask
	bpf_u_int32 net;			// IP addr
//	const u_char *packet;			// pointer to the packet
//	struct pcap_pkthdr header;		// packet header

	if (verbose)
		printf("Original uid is %d\teuid is %d\n", getuid(), geteuid());

	/* gain superuser priviledges if setuid */
	setuid(0);

	if (verbose)
		printf("New uid is %d\teuid is %d\n", getuid(), geteuid());

	if (getuid() != 0)
	{
		fprintf(stderr, "You must be superuser to run this program.\n");
		exit(1);
	}

	yohood = hash_create(MAXHOSTS, sizeof(struct host));
	assert(yohood);

	dev = pcap_lookupdev(errbuf);
	pcap_lookupnet(dev, &net, &mask, errbuf);

	if (verbose)
		printf("Sniffing on interface %s\n", dev);

	handle = pcap_open_live(dev, BUFSIZ, 1, 0, errbuf);
	pcap_compile(handle, &filter, filter_expr, 0, net);
	pcap_setfilter(handle, &filter);

//	packet = pcap_next(handle, &header);

	pcap_loop(handle, -1, yohoo_get_packet, NULL);

	pcap_close(handle);

	return 0;
}

char* mac_ntoa(char *buffer, unsigned char *addr)
{
	int i;
	char temp[4];

	assert(buffer);

	buffer[0] = '\0';

	for (i = 0; i < ETH_ALEN; i++)
	{
		sprintf(temp, "%02X", addr[i]);
		strcat(buffer, temp);
		if (i != (ETH_ALEN - 1))
			strcat(buffer, ":");
	}

	return buffer;
}

char *hwtype_ntoa(short hwtype)
{
	short hwtype_ = ntohs(hwtype);

	switch (hwtype_)
	{
		case 0x0001:	return "Ethernet";
		case 0x000B:	return "LocalTalk";
	}

	return "Unknown";
}

char *prot_ntoa(short prot)
{
	short prot_ = ntohs(prot);

	switch (prot_)
	{
		case 0x0800:	return "IP";
		default:	return "Unknown";
	}

	return "Unknown";
}
